import logging
import functools

from typing import Callable, Any

from active_directory import client, exceptions, _object_classes, _ldaps
from ppa import ui

from .records import (
    DomainController,
    User,
    UserData,
    Group,
    GroupData,
    Computer,
    ComputerData,
    OrganizationalUnit,
    OrganizationalUnitData,
)

logger = logging.getLogger(__name__)


def idempotent_create(func: Callable) -> Callable:
    @functools.wraps(func)
    def idempotent_create(*args, **kwargs) -> None:
        # Idempotent create functions should not care if the object already exists.
        try:
            return func(*args, **kwargs)
        except exceptions.ObjectExists as e:
            logger.debug("idempotent create function ignored the following exception")
            logger.exception(e)
            pass

    return idempotent_create


def idempotent_remove(func: Callable) -> Callable:
    @functools.wraps(func)
    def idempotent_remove(*args, **kwargs) -> None:
        # Idempotent remove functions should not care if the object does not exist.
        try:
            return func(*args, **kwargs)
        except exceptions.NoSuchObject as e:
            logger.debug("idempotent remove function ignored the following exception")
            logger.exception(e)
            pass

    return idempotent_remove


def ldap_function(func: Callable) -> Callable:
    @functools.wraps(func)
    def _fn(domain_controller: DomainController, *args, **kwargs) -> Any:
        # The opening of the connection should only happen if domain_controller isn't a SimpleLDAPObject yet.
        if not isinstance(domain_controller, _ldaps.ldap.ldapobject.LDAPObject):
            if missing_keys := [key for key in {"address", "domain", "username", "password"} if key not in domain_controller]:
                msg = f"Active Directory functions require address, domain, username, and password in the connection details. The following are missing: {', '.join(missing_keys)}."
                ui.output_error(msg)
                raise Exception(msg)
            if "port" not in domain_controller:
                domain_controller["port"] = 636
            try:
                with client.initialize(**domain_controller)() as open_connection:
                    return func(open_connection, *args, **kwargs)
            except exceptions.ServerDown as e:
                msg = f"Can't connect to {domain_controller['address']} over LDAPS on port 636"
                ui.output_error(msg)
                raise e
            except exceptions.InvalidCredentials as e:
                ui.output_error(e.args[0])
                raise e

        # Already a SimpleLDAPObject (decorated function calling another decorated functions).
        return func(domain_controller, *args, **kwargs)

    return _fn


def process_user(user: _object_classes.User) -> User:
    return UserData(user, inherited_properties=["is_disabled", "is_enabled", "is_expired", "is_locked", "password_age"]).__dict__


def process_group(group: _object_classes.Group) -> Group:
    group = group.__dict__
    # Rather than confuse people by storing members in both "member" and "member_range", simplify
    # things at the plugin level by simply copying "member_range" into "member" if it exists. The
    # "member_range" key only exists if the group has > 1500 members, & in this case the "member"
    # key is always empty. This is also why the Group class is the only one that inherits from
    # CastDataClass. We need to pass it a dictionary rather than an object, as it's easier to
    # fiddle the value this way.
    if member_range := group.pop("member_range", None):
        group["member"] = member_range
    return GroupData(**group).__dict__


def process_computer(computer: _object_classes.Computer) -> Computer:
    return ComputerData(computer).__dict__


def process_organizational_unit(ou: _object_classes.OrganizationalUnit) -> OrganizationalUnit:
    return OrganizationalUnitData(ou).__dict__


def bool_to_search_scope(value: bool) -> int:
    # Converts the one_level boolean argument to an int.
    #
    # True = 1 (ldap.SCOPE_ONELEVEL)
    # False = 2 (ldap.SCOPE_SUBTREE)
    return 1 if value is True else 2
